package ClusterFisher;

use strict;
use warnings;

# Ian Grieve
# September 2006
use Carp;
use Exporter 'import';
use Statistics::FET 'fishers_exact';
require Math::Pari;

#############################################################
# ClusterFisher module of PoGO.
# Gets the 2x2 contingency table parameters for each POSOC cluster. A Fisher's test will be carried out for the cluster by FisherExact and the outcome of this test (a Fisher's test raw p-value) will be returned to main script.
# The cells of the 2x2 contingency table will be referred to as follows:
# $a: The number of DE genes annotated to the Cluster, formerly known as $cluster_de_counter.
# $b: The number of non-DE genes annotated to the Cluster, formerly known as $cluster_nonde_counter.
# $c: The number of DE genes, of those with GO annotations, which are not annotated to the cluster (it is not possible for a DE gene without GO annotations to be annotated to the cluster). Formerly known as $de_notcluster_anonly.
# $d: All other genes. Formerly known as $noncluster_nonde_counter.
# These four values, along with a specification of 'one-tailed' or 'two-tailed' (as given in the PoGO config file) are passed to FisherExact module.
#############################################################

sub Contingency_tables
{
    my ($de_genes, $posoc_clusters, $transfile, $min_de_genes) = @_;
    my ($a, $b, $c, $d);
    my %cluster_contingency_table;

# This part of the module is adapted from fisher_test_generator.pl. The Fisher's Exact Test will be carried out in Perl rather than R, but the method of obtaining the contingency table parameters is essentially the same.

    my $not_annotated_count = &Find_de_noannotations ($de_genes, $transfile);
    
    open (POSOC_CLUSTERS, $posoc_clusters) || die print "Cannot open POSOC Clusters file $posoc_clusters in ClusterFisher.pm, Contingency_tables subroutine.\n";
    open (DE_GENELIST, $de_genes) || die print "Cannot open DE Genes file $de_genes in ClusterFisher.pm, Contingency_tables subroutine.";

    my @de_genelist = <DE_GENELIST>;
    my @posoc_clusters_tosplit = <POSOC_CLUSTERS>;
    my $posoc_clusters_tosplit = "@posoc_clusters_tosplit";
    my @posoc_clusters = split /-{45}/, $posoc_clusters_tosplit;

    my $posoc_cluster_genes = $posoc_clusters[2]; #'genes' here are Affy IDs. This section of the POSOC clustering output gives the Affy IDs annotated to each cluster.
    my @posoc_cluster_genes = split /Cluster\s/, $posoc_cluster_genes;
	
    my $total_genes_clusters_info = $posoc_cluster_genes[0];
    $total_genes_clusters_info =~ /(\d+) genes/;
    my $clusters_genes_total = $1;

    shift @posoc_cluster_genes;

    if ($de_genelist[0] !~ /_[a|s]t/)
    {
	shift (@de_genelist);
    } 

    foreach my $cluster (@posoc_cluster_genes)
    {
	$cluster =~ /^(\d+)\((\d+) genes\): \*?GO:\d+ ([^=]*)=/;
	my $cluster_number = $1;
	my $cluster_totalgenes = $2;
	my ($cluster_name_print, $cluster_name_discard) = split / score/, $3;
	$cluster_name_print =~ s/\s$//;
	my $cluster_id = $cluster_name_print."*".$cluster_number;
	my $cluster_de_counter = 0;
	my $de_notcluster_counter = 0;
	
	foreach my $de_gene (@de_genelist)
	{
	    $de_gene =~ /([^\s]*[a|s]t)\s*/;
	    my $de_gene_tomatch = $1;
	    if ($cluster =~ /$de_gene_tomatch/)
	    {
		$cluster_de_counter = $cluster_de_counter + 1;
		# Cluster_de_counter counts the number of differentially expressed genes in each cluster
	    }
	    
	    else 
	    {
		$de_notcluster_counter = $de_notcluster_counter + 1;
		# DE_notcluster_counter counts those differentially expressed genes that are not in the cluster being tested.
	    }
	    
	}

    if ($cluster_de_counter < $min_de_genes)
    # This if statement is used if a minimum number of DE genes to generate a p-value is selected on the command line.
    {
	next;
    }

    else
    {
#	$de_totalgenes = $cluster_de_counter + $de_notcluster_counter;

	$a = $cluster_de_counter;
	$b = $cluster_totalgenes - $cluster_de_counter;
	$c = $de_notcluster_counter - $not_annotated_count;
	$d = $clusters_genes_total - $a - $b - $c;
	
	my @current_cluster_contingency_table = ($a,$b,$c,$d);
	$cluster_contingency_table{$cluster_id} = \@current_cluster_contingency_table;
    }
    }

    my %return_fisher_pvalues = &Run_fisher_exact_tests(\%cluster_contingency_table);
    return %return_fisher_pvalues;
}

sub Find_de_noannotations
{
    # Find the count of DE genes that have no GO annotations in the transfile used by POSOC.
    my ($de_genes, $if_present_annotated) = @_;
    my $annotated_count = 0;
    my $not_annotated_count = 0;
 
    open (DE_GENES, $de_genes) || die print "Cannot open DE Genes file $de_genes in ClusterFisher.pm, find_de_noannotations subroutine";
    open (IF_PRESENT_ANNOTATED, $if_present_annotated) || die print "Cannot open Transfile $if_present_annotated in ClusterFisher.pm, find_de_noannotations subroutine";

    $/ = "\n";
    my @if_present_annotated = <IF_PRESENT_ANNOTATED>;
    my $if_present_annotated_tosearch = "@if_present_annotated";
    my $annotated_genes_total = $#if_present_annotated;
    my @de_genes = <DE_GENES>;

    while ($de_genes[0] !~ /_[a|s]t/)  
    {
	shift (@de_genes);
    }
    
    foreach my $de_gene (@de_genes)
    {
	chomp $de_gene;
	$de_gene =~ /([^\s]*)\s*/;
	my $de_gene_id = $1;

	if ($if_present_annotated_tosearch =~ /$de_gene_id/)
	{
	    # For each Affy ID in the list of differentially expressed Affy IDs, finds whether it is also found in the list of annotated Affy IDs (i.e. the transfile).
	    $annotated_count = $annotated_count + 1;
	}
	
	else 
	{
	    $not_annotated_count = $not_annotated_count + 1;
	}
    }

#    my $total_count_debug = $annotated_count + $not_annotated_count;
    
    return $not_annotated_count;
}

sub Run_fisher_exact_tests
{

    my $fisher_p_values = 'fisher_p_values.txt';
    open (FISHER, ">$fisher_p_values") || die print "Run_fisher_exact_tests subroutine in ClusterFisher.pm could not open $fisher_p_values\n";

    my (%cluster_contingency_tables) = %{$_[0]};
    my %fisher_return = ();
    my @cluster_names_to_test = keys %cluster_contingency_tables;

    foreach my $cluster_to_test (@cluster_names_to_test)
    {
	my $current_cluster_contingency_table_ref = $cluster_contingency_tables{$cluster_to_test};
	my @current_cluster_contingency_table = @$current_cluster_contingency_table_ref;
	my $a = $current_cluster_contingency_table[0];
	my $b = $current_cluster_contingency_table[1];
	my $c = $current_cluster_contingency_table[2];
	my $d = $current_cluster_contingency_table[3];  
	print FISHER "CLUSTER TO TEST\nNAME: $cluster_to_test\nA: $a\nB: $b\nC: $c\nD: $d\n";
	
	my $fs = 1;
	
	my $fisher_p_value = &fishers_exact($a, $b, $c, $d, $fs);
	print FISHER "FISHER P-VALUE: $fisher_p_value\n\n";
	my $fisher_p_value_rounded = $fisher_p_value;
#	my $fisher_p_value_rounded = sprintf("%.4f", $fisher_p_value); 
#       If you want rounded p-values, replace the 'my $fisher_p_value_rounded' line with the above. Change 4 to another number to get different numbers of decimal places. I have not enabled this by default because very low (i.e. significant) p-values come out as 0.
	
	if ($fisher_p_value_rounded > 1)
	{
	    $fisher_p_value_rounded = 1;
	}	

	$fisher_return{$cluster_to_test} = "$fisher_p_value_rounded";
    }
    
    close (FISHER);
    return %fisher_return;
}



1;

