#!/usr/local/bin/perl
#
# The new version of the PoGO scripts to reduce the required amount of user input and streamline the process.
# September 2006
# Ian Grieve

# The required files should be specified in the PoGO configuration file (the config file MUST be named as follows: PoGO_config.txt, or this script won't be able to find it and will die). These should be as follows:
# MTX2: The name of the MTX2 file that was used by POSOC to generate the clusters. This should be distilled from the CSV file obtained from the Affymetrix website (NetAffx resource) using affyfile_converter.pl if not already present.
# FULL: The full list of probesets (in the form of Affymetrix IDs, one per line) on the chip.
# DE: The list of probesets considered to be differentially expressed or otherwise of interest for comparison against the baseline (again, in the form of Affymetrix IDs, one per line).
# CLUSTERS: The clustering output produced by POSOC.

# Below this are the specified parameters for tests carried out by PoGO. These are as follows.
# FISHER_MIN: The minimum number of DE genes annotated to a cluster for it to be included in Fisher's testing and in the Benjamini-Hochberg multiple testing correction. This should be an integer. Default 0.
# PERMS: Number of permutations to be carried out in permutation testing. This should be an integer. Default 1000.
# BH_SIG: Benjamini-Hochberg significance threshold. Default 0.1 (10%).

# All of the files should be in the same directory as the script (the configuration file and the four data files specified above).

use strict;
use Modules::ClusterFisher;
use Modules::PermutationTests;
use Modules::BenjaminiHochberg;
use Data::Dumper;
use Math::Pari ();
use List::Util ();

# Reads PoGO_config.txt to establish the filenames of the input files and the parameters for the tests.

my $config = 'PoGO_config.txt';
open (CONFIG, $config) || die ("Could not find PoGO configuration file $config. Ensure that this file is present in the working directory and it is named correctly.\n");

my $fisherdebug = 'fisher_debug.txt';
open (FISHERDEBUG, ">$fisherdebug");

$/ = "====";
my @config = <CONFIG>;
shift (@config);
$/ = "\n";

my ($mtx2, $full, $de, $clusters, $fisher_min, $perms, $bh_sig);

foreach my $config_section (@config)
{
    my @config_line = split("\n", $config_section);
    shift (@config_line);
    
    my @config_line_a_desc = split("\: ", $config_line[0]);
    my @config_line_b_desc = split("\: ", $config_line[1]);
    my @config_line_c_desc = split("\: ", $config_line[2]);
    my @config_line_d_desc = split("\: ", $config_line[3]);
    
    if ($config_line_a_desc[0] eq 'MTX2')
    {
	$mtx2 = $config_line_a_desc[1];
	$full = $config_line_b_desc[1];
	$de = $config_line_c_desc[1];
	$clusters = $config_line_d_desc[1];
    }

    elsif ($config_line_a_desc[0] eq 'FISHER_MIN')
    {
	$fisher_min = $config_line_a_desc[1];
	$perms = $config_line_b_desc[1];
	$bh_sig = $config_line_c_desc[1];
    }
}


# Prepares and carries out Fisher's Exact Tests on each cluster above the minimum DE threshold.
print "Fisher's Tests In Progress\n";

my %clusters_fisher_out = ClusterFisher::Contingency_tables($de,$clusters,$mtx2,$fisher_min);
print FISHERDEBUG Dumper %clusters_fisher_out;

my @fisher_cluster_names_ids = keys(%clusters_fisher_out);

foreach my $fisher_hash_key (@fisher_cluster_names_ids)
{
    my @fisher_hash_key_split = split(/\*/, $fisher_hash_key);
}

close (FISHERDEBUG);

# Prepares and carries out Permutation Tests on each cluster above the minimum DE threshold.

# The required files and parameters are as follows:
# File 1: The list of all Affy IDs on the chip ($full, given in config file by FULL).
# File 2: The list of differentially expressed Affy IDs ($de, given in config file by DE).
# File 3: The POSOC clustering output for the chip ($clusters, given in config file by CLUSTERS).
# Also required is the parameter that specifies the number of permutations to carry out for each cluster ($perms, given in config file by PERMS),
# and the parameter that specifies the minimum number of DE genes in a cluster for it to be tested ($fisher_min, given in config file by FISHER_MIN).
print "Permutation Tests In Progress\n";
open (FULL_LIST, $full) || die("Cannot open $full in PoGO, Permutation Tests section.\n");
my @full_genes_list = <FULL_LIST>;

open (DE_LIST, $de) || die print "Cannot open $de in PoGO, Permutation Tests section.\n";
my @de_genes_list = <DE_LIST>;

my $randomized_list_length = scalar(@de_genes_list); # How many random genes from the FULL list to use in the permutation tests - needs to be equivalent to the number of DE genes in the experiment being analysed.

# The cluster Affy IDs will be matched against this random list and the number matching compared with that matching against the real DE genes list. This will be repeated for $perms permutations.

my $permutation_debug = 'permutation_debug.txt';
open (POSOC_CLUSTERS, $clusters) || die("Cannot open $clusters in PoGO, Permutation Tests section.\n");

my @posoc_clusters_tosplit = <POSOC_CLUSTERS>;
my $posoc_clusters_tosplit = "@posoc_clusters_tosplit";
my @posoc_clusters = split /-{45}/, $posoc_clusters_tosplit; # If the file format changes this line will probably have to be altered.

my $posoc_cluster_genes = $posoc_clusters[2]; # This part contains the Affy IDs that belong to each POSOC cluster.
my @posoc_cluster_genes = split/Cluster\s/, $posoc_cluster_genes; # Separate each Cluster into a different element of the array.
shift (@posoc_cluster_genes);
my $total_posoc_clusters = scalar(@posoc_cluster_genes);

for (my $iteration = 1; $iteration <= $perms; $iteration++) # Perform the matching function for the number of permutations specified by PERMS.
{
    my @current_iteration_randomized_list = PermutationTests::Randomize ($randomized_list_length, @full_genes_list);
    my $current_iteration_match_genes = PermutationTests::Cluster_affymatch ($permutation_debug, $iteration, \@posoc_cluster_genes, \@current_iteration_randomized_list);
}
	
close (FULL_LIST);


# Having matched each of the $perms permutations of random genes to the clusters, now do the same for the real DE Affy IDs list.

my @nonrandomized_de_list;

foreach my $de_list_affyid (@de_genes_list)
{
    chomp ($de_list_affyid);
    my @de_list_affyid_split = split(/\t/, $de_list_affyid);
    push (@nonrandomized_de_list, $de_list_affyid_split[0]);
}

my $de_counts = 'de_counts.txt';
my $de_match_genes = PermutationTests::Cluster_affymatch ($de_counts, '1', \@posoc_cluster_genes, \@nonrandomized_de_list);

my $vector_counts_out = 'vector_counts.txt';
open (VECTOR_COUNTS, ">>$vector_counts_out") || die ("Cannot open $vector_counts_out in PoGO, Permutation Tests section.\n");

our $perminput = 'permutation_debug.txt';
open (PERMUTATION_INPUT, $perminput) || die ("Cannot reopen $perminput in PoGO, Permutation Tests section.\n"); # This is the output file created by Cluster_affymatch using the $perms random sets of Affy IDs.

my @perminput = <PERMUTATION_INPUT>;
my $perminput_tosplit_byiteration = "@perminput";
my @perminput_split_byiteration = split /\#\# Iteration/, $perminput_tosplit_byiteration;
shift @perminput_split_byiteration;

# Create vectors of the number of matches corresponding to each cluster that the number of matches by the DE Affy IDs can be compared with. 
my $vector_iteration = 1;

while ($vector_iteration <= $total_posoc_clusters)
{
   my $vector_counts_out = PermutationTests::Vector_maker($vector_iteration, \@perminput_split_byiteration);
   $vector_iteration = $vector_iteration + 1;
} 
    
close (VECTOR_COUNTS);
close (PERMUTATION_INPUT);
close (DE_LIST);

# Carry out the comparison of the vector counts with the 'actual counts' from the DE Affy IDs list for each cluster. This is where the clusters are filtered for minimum no. of DE genes (using $fisher_min).
my $vector_counts_in = 'vector_counts.txt';
open (VECTOR_COUNTS_IN, $vector_counts_in) || die ("Cannot reopen $vector_counts_in in PoGO, Permutation Tests section.\n");
my @vector_counts_input = <VECTOR_COUNTS_IN>;

my $de_input = 'de_counts.txt';
open (DE_INPUT, $de_input) || die ("Cannot reopen $de_input in PoGO, Permutation Tests section.\n");
my @de_counts_in = <DE_INPUT>;

my %perm_test_results = PermutationTests::Vector_comparer($fisher_min, \@vector_counts_input, \@de_counts_in);
# The permutation results is a hash of arrays using Cluster No. as the Key. The array elements are as follows: 0 - Number of DE Affy IDs matching cluster; 1 - Probability that DE Affy IDs are overrepresented in the cluster (compared to null sets); 2 - Probability that DE Affy IDs are underrepresented in the cluster (compared to null sets).

close (VECTOR_COUNTS_IN);
close (DE_INPUT);

# These delete the debug and temp files used by the script. If you are having problems or want to see intermediate calculations carried out by PoGO, comment them out. But make sure you delete the file(s) before you run the script again, or you might get some strange results next time!
unlink 'fisher_debug.txt';
unlink 'permutation_debug.txt';
unlink 'de_counts.txt';
unlink 'fisher_p_values.txt';
unlink 'permutation_tests.txt';
unlink 'vector_counts.txt';

# Carry out Multiple Testing Correction & Output Results.

# The data that is required for output to the final output file is contained in the two hashes: %clusters_fisher_out, which contains the name of the cluster (cluster head GO term), the cluster number, and the Fisher's p-value for that cluster; and %perm_test_results, which contains the cluster number, the number of DE genes in the cluster, the 'DE over-represented' p-value and the 'DE under-represented' p-value.

# Carry out the multiple testing correction, using the Benjamini-Hochberg method. The FDR level specified in the config file BH_SIG ($bh_sig) is used. Outputs the clusters whose Fisher p-values are significant at the specified level.
print "Multiple Testing Correction In Progress\n";
my @bh_significant_clusters = BenjaminiHochberg::Sort_pvalues($bh_sig, \%clusters_fisher_out);
print "Clusters for which null hypothesis is rejected: @bh_significant_clusters\n\n"; # Cluster numbers of the Benjamini-Hochberg significant clusters.

my $results_output = 'PoGO_results.txt';
open (RESULTS_OUT, ">$results_output") || die ("Cannot open $results_output output file in PoGO, Results Outputting section.\n");
print RESULTS_OUT "Cluster Head\tCluster No.\tNo. DE Affy IDs\tPerms P(DE Overrep.)\tPerms P(DE Underrep.)\tFisher p-value\tBH Sig.\n";

# All of this process until the next comment is to sort the output into ascending order of Fisher's p-value.
my @all_clusters_details = keys(%clusters_fisher_out);
my @all_clusters_pvaluesnames;

foreach my $cluster_details (@all_clusters_details)
{
    my $current_cluster_pvalue = $clusters_fisher_out{$cluster_details};;
    my $current_cluster_details = $current_cluster_pvalue."\t".$cluster_details;
    push (@all_clusters_pvaluesnames, $current_cluster_details);
}

my @pvalues_sorted = sort { (split /\t/, $a)[0] <=> (split /\t/, $b) [0] } @all_clusters_pvaluesnames; # This is supposed to sort the clusters...
my @pvalues_resorted;

foreach my $cluster_details (@pvalues_sorted)
{
    my @cluster_details_split = split(/\t/, $cluster_details);
    push (@pvalues_resorted, $cluster_details_split[1]);
}

# Collates all of the data of interest for each cluster for printing to the final output file, 'PoGO_results.txt'.
print "Printing Results\n";
foreach my $cluster_details (@pvalues_resorted)
{
    my @cluster_details_split = split(/\*/, $cluster_details);
    my $current_cluster_name = $cluster_details_split[0];
    my $current_cluster_number = $cluster_details_split[1];
    my $current_cluster_fisher_pvalue = $clusters_fisher_out{$cluster_details};
    my $current_cluster_permtest_results_ref = $perm_test_results{$current_cluster_number};
    my $cluster_bh_sig_bool = 0;

    foreach my $is_cluster_bh_sig (@bh_significant_clusters)
    {
	if ($current_cluster_number == $is_cluster_bh_sig)
	{
	    $cluster_bh_sig_bool = 1;
	}
    }
    
    print RESULTS_OUT "$current_cluster_name\t$current_cluster_number\t$$current_cluster_permtest_results_ref[0]\t$$current_cluster_permtest_results_ref[1]\t$$current_cluster_permtest_results_ref[2]\t$current_cluster_fisher_pvalue\t$cluster_bh_sig_bool\n"; # Tab delimited results outputting file, sorted by p-value.
}

print "PoGO Process Complete.\n";
















